"""
service.service

MacOS System and GUI domain services.
"""

import logging
import os
from pathlib import Path

from . import launchctl


__all__ = ["locate", "Service"]


logger = logging.getLogger(__name__)


class Service:
    """A LaunchAgent or LaunchDaemon service.

    :param path: The path to the service file.
    """

    def __init__(self, path: Path):
        self._path = path

    @property
    def domain(self) -> str:
        """The service domain."""
        return launchctl.DOMAIN_SYS if os.getenv("SUDO_USER") else f"{launchctl.DOMAIN_GUI}/{os.geteuid()}"

    @property
    def file(self) -> str:
        """The absolute path to the service file."""
        return str(self._path.absolute())

    @property
    def id(self) -> str:
        """The service ID in the system domain."""
        return "/".join([self.domain, self.name]) if os.getenv("SUDO_USER") else ""

    @property
    def name(self) -> str:
        """The service name (service file name without extension)."""
        return self._path.stem

    @property
    def path(self) -> Path:
        """The path to the service file."""
        return self._path

    def validate(self) -> None:
        """Validate the service.

        A service is considered valid if it is part of the active domain and is not a macOS system service.

        :raises RuntimeError: When the service is not valid.
        """
        if self.domain == launchctl.DOMAIN_SYS:
            if self.file.startswith("/System"):
                raise RuntimeError(f"{self.name} is a macOS system service")

            if self.file.startswith("/Users"):
                raise RuntimeError(f"{self.name} is not in the {self.domain} domain")
        else:
            if not self.file.startswith("/Users"):
                raise RuntimeError(f"{self.name} is not in the {self.domain} domain")


def locate(name: str, reverse_domains: list[str]) -> Service:
    """Locate a service.

    If an absolute or relative path is part of `name` that path is used to find the service. If a path is not present
    all directories containing services for the current domain will be searched.

    :param name: The service name, with optional absolute/relative path and file extension.
    :param reverse_domains: A list of reverse domains to prepend to the service name.

    :raises ValueError: When a service is not found or a service name without path and/or domain is provided and there
    no reverse domains are configured.
    """
    logger.debug('Locating service "%s"', name)
    original_name = name
    service_path = None

    if not name.endswith(".plist"):
        name = f"{name}.plist"

    path = Path(name)

    logger.debug("Generating potential file paths")

    if len(path.parts) > 1:
        file_paths = [path.expanduser().absolute()]
    else:
        if len(path.suffixes) > 1:
            file_names = [path.name]
        else:
            if reverse_domains:
                file_names = [f"{rd}.{path.name}" for rd in reverse_domains]
            else:
                raise ValueError("No reverse domains configured")

        file_paths = [p.joinpath(n) for p in get_paths() for n in file_names]

    for file_path in file_paths:
        logger.debug('Trying "%s"', file_path)

        if file_path.is_file():
            service_path = file_path
            break

    if not service_path:
        raise ValueError(f'Service "{original_name}" not found')

    logger.debug('Service found, using "%s"', service_path)
    service = Service(service_path)

    logger.debug("Validating service")
    service.validate()

    return service


def get_paths() -> list[Path]:
    """Get service paths for the active domain.

    :raises ValueError: When no service paths are found.
    """
    logger.debug("Identifying service paths")

    base_paths = ["/", "/System"] if os.getenv("SUDO_USER") else [Path.home()]
    service_paths = []

    for base in base_paths:
        for path in ["Library/LaunchAgents", "Library/LaunchDaemons"]:
            service_path = Path(base, path)

            if service_path.is_dir():
                service_paths.append(service_path)

    if not service_paths:
        raise ValueError("No service paths found")

    return service_paths



"""
service.cli

The command-line interface for service
"""

import logging
import os
from pathlib import Path
import platform
import typing as t

import click
from clickext import ClickextCommand, ClickextGroup, config_option, verbose_option

from . import launchctl
from .service import locate, Service


MACOS_MIN_VERSION = 12.0
CONFIG_FILE = Path(f'~{os.getenv("SUDO_USER", "")}/.config/service.toml').expanduser()


logger = logging.getLogger(__package__)


def get_reverse_domains(data: t.Optional[dict[str, list[str]]]) -> list[str]:
    """Build reverse domains from configuration file data.

    :param data: The parsed configuration file data.
    """
    if data is None or "reverse-domains" not in data:
        reverse_domains = []
    else:
        reverse_domains = data["reverse-domains"]

        if not isinstance(reverse_domains, list):
            logger.warning('Invalid configuration file. "reverse-domains" must be a list.')
            reverse_domains = []

    logger.debug("Configured with %s reverse domains", len(reverse_domains))

    return reverse_domains


def get_service(ctx: click.Context, param: click.Parameter, value: str) -> None:  # pylint: disable=unused-argument
    """Get the target service and store it on `ctx.obj`.

    :param ctx: The current click execution context.
    :param param: The parameter that triggered the callback.
    :param value: The parameter value.
    """
    reverse_domains: list[str] = ctx.obj.copy()
    ctx.obj = locate(value, reverse_domains)


@click.group(cls=ClickextGroup, global_opts=["config", "verbose"], shared_params=["name"])
@click.argument("name", nargs=1, callback=get_service, expose_value=False, type=click.STRING)
@click.version_option(package_name="py_service")
@config_option(CONFIG_FILE, processor=get_reverse_domains)
@verbose_option(logger)
def cli() -> None:
    """Extremely basic launchctl wrapper for macOS."""
    logger.debug("%s started", __package__)
    verify_platform()


@cli.command(cls=ClickextCommand)
@click.pass_obj
def disable(service: Service) -> None:
    """Disable a service (system domain only)."""
    launchctl.change_state(service, enable=False)
    logger.info("%s disabled", service.name)


@cli.command(cls=ClickextCommand)
@click.pass_obj
def enable(service: Service) -> None:
    """Enable a service (system domain only)."""
    launchctl.change_state(service, enable=True)
    logger.info("%s enabled", service.name)


@cli.command(cls=ClickextCommand)
@click.pass_obj
def restart(service: Service) -> None:
    """Restart a service."""
    launchctl.boot(service, run=False)
    launchctl.boot(service, run=True)
    logger.info("%s restarted", service.name)


@cli.command(cls=ClickextCommand)
@click.option(
    "--enable",
    "-e",
    "enable_service",
    is_flag=True,
    default=False,
    help="Enable sevice before starting (system domain only).",
)
@click.pass_obj
def start(service: Service, enable_service: bool) -> None:
    """Start a service."""
    if enable_service:
        launchctl.change_state(service, enable=True)

    launchctl.boot(service, run=True)
    logger.info("%s %sstarted", service.name, "enabled and " if enable_service else "")


@cli.command(cls=ClickextCommand)
@click.option(
    "--disable",
    "-d",
    "disable_service",
    is_flag=True,
    default=False,
    help="Disable service after stopping (system domain only).",
)
@click.pass_obj
def stop(service: Service, disable_service: bool) -> None:
    """Stop a service."""
    launchctl.boot(service, run=False)

    if disable_service:
        launchctl.change_state(service, enable=False)

    logger.info("%s stopped%s", service.name, " and disabled" if disable_service else "")


def verify_platform() -> None:
    """Verify the platform is supported.

    :raises RuntimeError: When the platform or platform version is not supported.
    """
    logger.debug("Checking platform")

    if platform.system() != "Darwin":
        raise click.ClickException(f"{__package__} requires macOS")

    macos_version = platform.mac_ver()[0]
    macos_version = float(".".join(macos_version.split(".")[:2]))

    if macos_version < MACOS_MIN_VERSION:
        raise click.ClickException(f"{__package__} requires macOS {MACOS_MIN_VERSION} or higher")



"""
service.launchctl

An interface for constructing and executing launchctl commands.
"""

from __future__ import annotations
import logging
import subprocess
import typing as t

if t.TYPE_CHECKING:
    from .service import Service


__all__ = ["DOMAIN_GUI", "DOMAIN_SYS", "boot", "change_state"]


DOMAIN_GUI = "gui"
DOMAIN_SYS = "system"

ERROR_GUI_ALREADY_STARTED = 5
ERROR_GUI_ALREADY_STOPPED = 5
ERROR_SIP = 150
ERROR_SYS_ALREADY_STARTED = 37
ERROR_SYS_ALREADY_STOPPED = 113


logger = logging.getLogger(__name__)


def _execute(subcommand: str, *args: str) -> None:
    """Construct and execute a launchctl command.

    :param subcommand: The launchctl subcommand to run
    :param args: The arguments for the subcommand
    """
    cmd = ["launchctl", subcommand, *args]

    logger.debug('Calling launchctl with command "%s"', " ".join(cmd))
    subprocess.run(cmd, check=True, capture_output=True)


def boot(service: Service, run: bool = False) -> None:
    """Start or stop a service.

    :param service: The service to modify.
    :param run: Whether to run (start) the service.

    :raises RuntimeError: When the service is already in the target state, runtime state change is prevented by SIP, or
    changing the runtime state fails.
    """
    subcmd, action, current_state = ("bootstrap", "start", "started") if run else ("bootout", "stop", "stopped")

    logger.debug("Changing service runtime state: %s (%s)", service.name, action)

    try:
        _execute(subcmd, service.domain, service.file)
    except subprocess.CalledProcessError as exc:
        if exc.returncode in [
            ERROR_GUI_ALREADY_STARTED,
            ERROR_GUI_ALREADY_STOPPED,
            ERROR_SYS_ALREADY_STARTED,
            ERROR_SYS_ALREADY_STOPPED,
        ]:
            msg = f"{service.name} is already {current_state}"
        else:
            reason = " due to SIP" if exc.returncode == ERROR_SIP else ""
            msg = f"Failed to {action} {service.name}{reason}"

        raise RuntimeError(msg) from exc


def change_state(service: Service, enable: bool = False) -> None:
    """Change service state (enable/disble).

    :param service: The service to target.
    :param enable: Whether the service should be enabled.

    :raises ValueError: When an unknown service state is specified.
    :raises RuntimeError: When the service state cannot be changed or changing the service state fails.
    """
    subcmd = "enable" if enable else "disable"

    logger.debug("Changing service state: %s (%s)", service.name, subcmd)

    if service.domain != DOMAIN_SYS:
        raise RuntimeError(f'Cannot change service state in the "{service.domain}" domain')

    try:
        _execute(subcmd, service.id)
    except subprocess.CalledProcessError as exc:
        raise RuntimeError(f"Failed to {subcmd} {service.name}") from exc
